from fantasynames.language import Language

# This file test to make sure capitalization and special transformation
# rules are applied as expected.


class TestLanguage(Language):
    @classmethod
    def _name1_male(cls) -> str:
        return "bob*y"

    @classmethod
    def _name1_female(cls) -> str:
        return "jack*y"

    @classmethod
    def _name2(cls) -> str:
        return "the brave d'blanche"


def test_male_name():
    name = TestLanguage.name("male")
    assert name == "Bobby the Brave d'Blanche"


def test_female_name():
    name = TestLanguage.name("female")
    assert name == "Jacky the Brave d'Blanche"



from fantasynames.data import french_data
import random
from fantasynames.language import Language


class French(Language):
    transformations = french_data["transformations"]

    @classmethod
    def _name1_male(cls) -> str:
        cols = [french_data["name1_col1"], french_data["name1_col2"]]
        return cls._name_from_lists(cols)

    @classmethod
    def _name1_female(cls) -> str:
        name = cls._name1_male() + cls._name_from_lists(
            [french_data["name1_female_suffixes"]]
        )
        return name

    @classmethod
    def _name2(cls) -> str:
        cols = [french_data["name2_col1"], french_data["name2_col2"]]
        name = cls._name_from_lists(cols)
        # 50% chance to use "name1 d' name2" or "name1 du name2" format
        if random.random() * 100 <= 30:
            prefix = random.choice(french_data["name2_prefixes"])
            name = prefix + name
        return name


french = French.name



from fantasynames.language import Language

# This file tests to make sure the basic process for name generation
# works as expected.

name1_col1 = ["lY"]
name1_col2 = ["n"]
name1_female_suffixes = ["a"]
name2_col1 = ["Ux"]
name2_col2 = ["tUn"]

test_tranformations = [
    {"input": "Y", "outputs": ["e"]},
    {"input": "U", "outputs": ["o"]},
]


class TestLanguage(Language):
    transformations = test_tranformations

    @classmethod
    def _name1_male(cls) -> str:
        cols = [name1_col1, name1_col2]
        return cls._name_from_lists(cols)

    @classmethod
    def _name1_female(cls) -> str:
        name = cls._name1_male() + cls._name_from_lists([name1_female_suffixes])
        return name

    @classmethod
    def _name2(cls) -> str:
        cols = [name2_col1, name2_col2]
        name = cls._name_from_lists(cols)
        return name


def test_name():
    name = TestLanguage.name()
    assert name == "Lena Oxton" or name == "Len Oxton"


def test_male_name():
    name = TestLanguage.name("male")
    assert name == "Len Oxton"


def test_female_name():
    name = TestLanguage.name("female")
    assert name == "Lena Oxton"



from fantasynames.data import dwarf_data, compound_tables
from fantasynames.language import Language
import random


class Dwarf(Language):
    transformations = dwarf_data["transformations"]

    @classmethod
    def _name1_male(cls) -> str:
        cols = [dwarf_data["name1_col1"], dwarf_data["name1_col2"]]
        return cls._name_from_lists(cols)

    @classmethod
    def _name1_female(cls) -> str:
        # 50% to add "-a" to a male name
        if random.random() * 100 < 50:
            name = cls._name1_male() + "a"
        # 50% chance to use female col for second half of name
        else:
            cols = [dwarf_data["name1_col1"], dwarf_data["name1_col2_female"]]
            name = cls._name_from_lists(cols)
        return name

    @classmethod
    def _name2(cls) -> str:
        cols = [compound_tables["mountain_col1"], compound_tables["mountain_col2"]]
        name = cls._name_from_lists(cols)
        return name


dwarf = Dwarf.name



from fantasynames.data import elf_data, compound_tables
from fantasynames.language import Language
import random


class Elf(Language):
    transformations = elf_data["transformations"]

    @classmethod
    def _name1_male(cls) -> str:
        cols = [elf_data["name1_col1"], elf_data["name1_col2"]]
        return cls._name_from_lists(cols)

    @classmethod
    def _name1_female(cls) -> str:
        name = cls._name1_male() + cls._name_from_lists(
            [elf_data["name1_female_suffixes"]]
        )
        return name

    @classmethod
    def _name2(cls) -> str:
        # 50% chance of elf-y surname or nature-y surname
        if random.random() * 100 < 50:
            cols = [compound_tables["nature_col1"], compound_tables["nature_col2"]]
        else:
            cols = [elf_data["name1_col1"], elf_data["name2_col2"]]

        name = cls._name_from_lists(cols)
        return name


elf = Elf.name



from fantasynames.data import human_data, compound_tables
from fantasynames.language import Language
from fantasynames.anglo import Anglo
from fantasynames.french import French
import random


class Human(Language):
    @classmethod
    def _name1_male(cls) -> str:
        # 80% chance anglo name, 20% chance french
        if random.random() * 100 <= 20:
            name = French._name1("male")
        else:
            name = Anglo._name1("male")
        return name

    @classmethod
    def _name1_female(cls) -> str:
        # 80% chance anglo name, 20% chance french
        if random.random() * 100 <= 20:
            name = French._name1("female")
        else:
            name = Anglo._name1("female")
        return name

    @classmethod
    def _name2(cls) -> str:
        prob = random.random() * 100
        # 20% chance of anglo surname
        if prob <= 20:
            name = Anglo._name2()
        # 20% chance of french surname
        elif prob <= 40:
            name = French._name2()
        # 60% chance surname is cobbled together from various tables
        else:
            col1 = random.choice(
                [
                    compound_tables["mountain_col1"],
                    compound_tables["nature_col1"],
                    compound_tables["generic_col1"],
                    compound_tables["generic_col1"],
                ]
            )
            col2 = random.choice(
                [
                    compound_tables["mountain_col2"],
                    compound_tables["nature_col2"],
                    compound_tables["generic_col2"],
                    compound_tables["generic_col2"],
                    human_data["name2_col2"],
                ]
            )
            name = cls._name_from_lists([col1, col2])

        return name


human = Human.name



from fantasynames.data import human_data
from fantasynames.language import Language
import random


class Anglo(Language):
    transformations = human_data["transformations"]

    @classmethod
    def _name1_male(cls) -> str:
        # 50% chance of just using one col (aka one syllable)
        if random.random() * 100 < 50:
            cols = [human_data["name1_col1"]]
        else:
            cols = [human_data["name1_col1"], human_data["name1_col2"]]
        return cls._name_from_lists(cols)

    @classmethod
    def _name1_female(cls) -> str:
        name = cls._name1_male() + cls._name_from_lists([human_data["name1_suffixes"]])
        return name

    @classmethod
    def _name2(cls) -> str:
        cols = [human_data["name2_col1"], human_data["name2_col2"]]
        name = cls._name_from_lists(cols)
        # 50% chance to use "name1 of name2" format
        if random.random() * 100 < 50:
            return "of " + name
        else:
            return name


anglo = Anglo.name



from abc import ABC, abstractmethod
import typing as t
import random


class Language(ABC):
    transformations: t.List[t.Dict] = []

    @classmethod
    def name(cls, gender: str = "any") -> str:
        """
        Returns a randomly generated first and last name. This is the only method
        that we intend to be "publicly" used outside of this package.
        """
        name = cls._name1(gender) + " " + cls._name2()
        transformed_name = cls._transform(name)
        caps_name = cls._capitalize(transformed_name)
        return caps_name

    @classmethod
    def _name1(cls, gender: str = "any") -> str:
        """
        Returns a randomly generated first name with variable gender.
        By default, randomly chooses either masculine or feminine.
        """
        if gender == "female":
            name = cls._name1_female()
        elif gender == "male":
            name = cls._name1_male()
        elif gender == "any":
            name = cls._name1_any()
        else:
            msg = (
                "Valid string parameters for name generating functions are "
                '"female", "male", or "any". The given value was the '
                f'{type(gender).__name__} "{gender}".'
            )
            raise ValueError(msg)

        return name

    @classmethod
    @abstractmethod
    def _name1_female(cls) -> str:
        """
        Expected to return a random 'feminine' first name. Must be overridden.
        """
        pass

    @classmethod
    @abstractmethod
    def _name1_male(cls) -> str:
        """
        Expected to return a random 'masculine' first name. Must be overridden.
        """
        pass

    @classmethod
    def _name1_any(cls) -> str:
        """
        Randomly returns either a masculine or feminine first name (50/50 chance).
        """
        if random.random() * 100 < 50:
            return cls._name1_male()
        else:
            return cls._name1_female()

    @classmethod
    @abstractmethod
    def _name2(cls) -> str:
        """
        Expected to return a random surname. Must be overridden.
        """
        pass

    @classmethod
    def _transform(cls, name: str) -> str:
        """
        Given a string, loops over each char. If 'input' is present in the
        cls.transformations dict, "replaces" it with a randomly
        selected character from the dict. Returns a new string with the newly
        replaced chars. (No actual mutations performed.)

        Also has the following special rules for special "input" characters:
            - *: doubles previous char if not preceeded by a CVC pattern
        """
        new_string = ""
        for char in name:
            new_char = char
            # "*" doubles previous char if not preceeded by a CVC pattern
            if char == "*":
                if double_consonant(new_string):
                    new_char = new_string[-1]
                else:
                    new_char = ""
            # checks is character is in given transformations
            for transformation in cls.transformations:
                if char == transformation["input"]:
                    new_char = random.choice(transformation["outputs"])
                    break
            new_string += new_char
        return new_string

    @classmethod
    def _capitalize(cls, name: str) -> str:
        caps_list = []
        for word in name.split():
            # leave particles lowercase
            if word in ["of", "du", "del", "de", "la", "von", "the"]:
                caps_list.append(word)
            # handle d'
            elif word[:2] == "d'":
                fragment = word[2:].capitalize()
                caps_list.append("d'" + fragment)
            # capitalize the rest
            else:
                caps_list.append(word.capitalize())
        return " ".join(caps_list)

    @classmethod
    def _name_from_lists(cls, lists: t.List[t.List[str]]) -> str:
        """
        Given one or more list of strings, randomly selects a string from each
        and returns the concatenated result.
        """
        name = ""
        for string_list in lists:
            name += random.choice(string_list)
        return name


# ----------------------
#        Helpers
# ----------------------


def double_consonant(string: str) -> bool:
    # the "double consonat rule" means that if the chars preceeding a special character
    # ("*" in our case) match a consonant-vowel-consonant pattern, we double the
    # final consonant. This helper function tests for that pattern.
    # Examples if true: bob*y -> bobby, rin*e -> rinne
    # Examples if false: rick*y -> ricky, wood*y -> woody
    if len(string) < 3:
        return True

    vowels = "aeiou"
    prev_chars = string[-3:]
    pattern = ""
    for char in prev_chars:
        if char in vowels:
            pattern += "V"
        else:
            pattern += "C"
    return pattern == "CVC"
